/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is ForumZilla 0.1.1.
 *
 * The Initial Developer of the Original Code is Zapogee, Inc.
 * Portions created by Zapogee, Inc. are Copyright (C) 2001
 * Zapogee, Inc.  All Rights Reserved.
 *
 * Contributor(s):
 *   Myk Melez, <myk@melez.com>, original author
 */


//****************************************************************************//
// constants and global variables

var feedService;

// Handy conversion values.
const HOURS_TO_MINUTES = 60;
const MINUTES_TO_SECONDS = 60;
const SECONDS_TO_MILLISECONDS = 1000;
const MINUTES_TO_MILLISECONDS = MINUTES_TO_SECONDS * SECONDS_TO_MILLISECONDS;
const HOURS_TO_MILLISECONDS = HOURS_TO_MINUTES * MINUTES_TO_MILLISECONDS;

const MESSAGE_TEMPLATE = "\n\
<html>\n\
  <head>\n\
    <title>%TITLE%</title>\n\
    <style type=\"text/css\">\n\
      body {\n\
        margin: 0;\n\
        border: none;\n\
        padding: 0;\n\
      }\n\
      %STYLE%\n\
    </style>\n\
  </head>\n\
  <body>\n\
    %CONTENT_TEMPLATE%\n\
  </body>\n\
</html>\n\
";

const REMOTE_CONTENT_TEMPLATE = "\n\
    <iframe src=\"%URL%\">\n\
      %DESCRIPTION%\n\
    </iframe>\n\
";

const REMOTE_STYLE = "\n\
      iframe {\n\
        margin: 0;\n\
        border: none;\n\
        padding: 0;\n\
        position: fixed;\n\
        top: 0;\n\
        right: 0;\n\
        bottom: 0;\n\
        left: 0;\n\
      }\n\
";

const LOCAL_CONTENT_TEMPLATE = "\n\
    <div id=\"content\">\n\
      %CONTENT%\n\
    </div>\n\
";

const LOCAL_STYLE = "\n\
      #content {\n\
        margin: 0;\n\
        border: none;\n\
        padding: 0;\n\
      }\n\
";


//****************************************************************************//
// utility functions

function fromUnicode(str, charset) {
    var unicodeConverter =
      Components
        .classes["@mozilla.org/intl/scriptableunicodeconverter"]
          .createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
    unicodeConverter.charset = charset;
    return unicodeConverter.ConvertFromUnicode(str);
}

function toUtf8(str) {
    var unicodeConverter =
      Components
        .classes["@mozilla.org/intl/scriptableunicodeconverter"]
          .createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
    unicodeConverter.charset = "UTF-8";
    return unicodeConverter.ConvertFromUnicode(str);
}

function mimeEncode(str, charset) {  
    try {
        var mimeEncoder = Components
          .classes["@mozilla.org/messenger/mimeconverter;1"]
            .getService(Components.interfaces.nsIMimeConverter);
    } catch(e) {
        // nsIMimeConverter only available in post-2004-08-02 branch builds.
        return str;
    }

    // Try to encode using native charset; failing that, try encoding as UTF-8.
    var newStr;
    try {
        newStr = mimeEncoder.encodeMimePartIIStr(str, false, charset, 9, 72);
    }
    catch (e) {
        debug("couldn't encode as " + charset + ": " + e);
        newStr = str;
    }
    return newStr;
}

function W3CToIETFDate(dateString) {
    // Converts a W3C-DTF (subset of ISO 8601) date string to an IETF date string.
    // W3C-DTF is described in this note: http://www.w3.org/TR/NOTE-datetime
    // IETF is obtained via the Date object's toUTCString() method.  The object's
    // toString() method is insufficient because it spells out timezones on Win32
    // (f.e. "Pacific Standard Time" instead of "PST"), which Mail doesn't grok.
    // For info, see http://lxr.mozilla.org/mozilla/source/js/src/jsdate.c#1526.
  
    var parts = dateString.match(/(\d\d\d\d)(-(\d\d))?(-(\d\d))?(T(\d\d):(\d\d)(:(\d\d)(\.(\d+))?)?(Z|([+-])(\d\d):(\d\d))?)?/);
    //debug("date parts: " + parts);
  
    // Here's an example of a W3C-DTF date string and what .match returns for it.
    //  date: 2003-05-30T11:18:50.345-08:00
    //  date.match returns array values:
    //   0: 2003-05-30T11:18:50-08:00,
    //   1: 2003,
    //   2: -05,
    //   3: 05,
    //   4: -30,
    //   5: 30,
    //   6: T11:18:50-08:00,
    //   7: 11,
    //   8: 18,
    //   9: :50,
    //   10: 50,
    //   11: .345,
    //   12: 345,
    //   13: -08:00,
    //   14: -,
    //   15: 08,
    //   16: 00
  
    // Create a Date object from the date parts.  Note that the Date object
    // apparently can't deal with empty string parameters in lieu of numbers,
    // so optional values (like hours, minutes, seconds, and milliseconds)
    // must be forced to be numbers.
    var date = new Date(parts[1], parts[3]-1, parts[5], parts[7] || 0,
                        parts[8] || 0, parts[10] || 0, parts[12] || 0);
  
    // We now have a value that the Date object thinks is in the local timezone
    // but which actually represents the date/time in the remote timezone
    // (f.e. the value was "10:00 EST", and we have converted it to "10:00 PST"
    // instead of "07:00 PST").  We need to correct that.  To do so, we're going
    // to add the offset between the remote timezone and UTC (to convert the value
    // to UTC), then add the offset between UTC and the local timezone (to convert
    // the value to the local timezone).
  
    // Ironically, W3C-DTF gives us the offset between UTC and the remote timezone
    // rather than the other way around, while the getTimezoneOffset() method
    // of a Date object gives us the offset between the local timezone and UTC
    // rather than the other way around.  Both of these are the additive inverse
    // (i.e. -x for x) of what we want, so we have to invert them to use them
    // by multipying by -1
    // (f.e. if "the offset between UTC and the remote timezone" is -5 hours,
    // then "the offset between the remote timezone and UTC" is -5*-1 = 5 hours).
  
    // Note that if the timezone portion of the date/time string is absent
    // (which violates W3C-DTF, although ISO 8601 allows it), we assume the value
    // to be in UTC.
  
    // The offset between the remote timezone and UTC in milliseconds.
    var remoteToUTCOffset = 0;
    if (parts[13] && parts[13] != "Z") {
        var direction = (parts[14] == "+" ? 1 : -1);
        if (parts[15])
            remoteToUTCOffset += direction * parts[15] * HOURS_TO_MILLISECONDS;
        if (parts[16])
            remoteToUTCOffset += direction * parts[16] * MINUTES_TO_MILLISECONDS;
    }
    remoteToUTCOffset = remoteToUTCOffset * -1; // invert it
    //debug("date remote to UTC offset: " + remoteToUTCOffset);
  
    // The offset between UTC and the local timezone in milliseconds.
    var UTCToLocalOffset = date.getTimezoneOffset() * MINUTES_TO_MILLISECONDS;
    UTCToLocalOffset = UTCToLocalOffset * -1; // invert it
    //debug("date UTC to local offset: " + UTCToLocalOffset);
  
    date.setTime(date.getTime() + remoteToUTCOffset + UTCToLocalOffset);
  
    //debug("date string: " + date.toUTCString());
  
    return date.toUTCString();
}


//****************************************************************************//
// nsFeedItem component

function FeedItem() {}

FeedItem.interfaces = [Components.interfaces.nsIFeedItem,
                       Components.interfaces.nsIFileSpec,
                       Components.interfaces.nsIMsgCopyServiceListener,
                       Components.interfaces.nsISupports];

FeedItem.prototype.QueryInterface = function (iid) {
    for ( var i=0 ; i<FeedItem.interfaces.length ; i++) {
        if (iid.equals(FeedItem.interfaces[i]))
            return this;
    }

    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
}


//****************************************************************************//
// private methods

FeedItem.prototype.debug = function(txt) {
    debug(this.identity + " " + txt);
}


//****************************************************************************//
// nsIFeedItem interface

FeedItem.prototype._id = null;
FeedItem.prototype.__defineGetter__('id', function() { return this._id });
FeedItem.prototype.__defineSetter__('id', function(newval) { this._id = newval; });

FeedItem.prototype._feed = null;
FeedItem.prototype.__defineGetter__('feed', function() { return this._feed });
FeedItem.prototype.__defineSetter__('feed', function(newval) { this._feed = newval; });

FeedItem.prototype._url = null;
FeedItem.prototype.__defineGetter__('url', function() { return this._url });
FeedItem.prototype.__defineSetter__('url', function(newval) {
    var uri =
      Components
        .classes["@mozilla.org/network/standard-url;1"]
            .getService(Components.interfaces["nsIStandardURL"]);
    uri.init(1, 80, newval, null, null);
    uri = uri.QueryInterface(Components.interfaces.nsIURI);
    this._url = uri.spec;
});

FeedItem.prototype._description = null;
FeedItem.prototype.__defineGetter__('description', function() { return this._description });
FeedItem.prototype.__defineSetter__('description', function(newval) { this._description = newval; });

FeedItem.prototype._title = "(no subject)";
FeedItem.prototype.__defineGetter__('title', function() { return this._title });
FeedItem.prototype.__defineSetter__('title', function(newval) { this._title = newval; });

FeedItem.prototype._author = "anonymous";
FeedItem.prototype.__defineGetter__('author', function() { return this._author });
FeedItem.prototype.__defineSetter__('author', function(newval) { this._author = newval; });

FeedItem.prototype._content = null;
FeedItem.prototype.__defineGetter__('content', function() { return this._content; });
FeedItem.prototype.__defineSetter__('content', function(newval) { this._content = newval; });

// XXX Determine and stick to one consistent date format for this value.
// Note that we can't pass JS Date objects through XPCOM, so we have to make
// this a parsable date string or Unix-style "seconds since 1969" integer.
FeedItem.prototype._date = new Date().toString();
FeedItem.prototype.__defineGetter__('date', function() { return this._date; });
FeedItem.prototype.__defineSetter__('date', function(newval) { this._date = newval; });

// A human-readable string that identifies the item; used in debug statements.
FeedItem.prototype.__defineGetter__('identity', function() {
    return this.feed.title + ": " + this.title + " (" + this.id + ")";
});

FeedItem.prototype.__defineGetter__('charset', function() {
    return this.feed.charset || 'UTF-8';
});

FeedItem.prototype.__defineGetter__('messageID', function() {
    // XXX Make this conform to the message ID spec.
  
    var messageID = this.id || this.url || this.title;
  
    // Escape occurrences of message ID meta characters <, >, and @.
    messageID.replace(/</g, "%3C");
    messageID.replace(/>/g, "%3E");
    messageID.replace(/@/g, "%40");
  
    messageID = messageID + "@" + "localhost.localdomain";
  
    return messageID;
});

// XXX This should happen in the constructor automatically.
FeedItem.prototype.__defineSetter__('valid',
function(valid) {
    if (valid) {
        this.debug("validating");
    
        var ds = feedService.itemsDS;
        var resource = rdf.GetResource(this.url || ("urn:" + this.id));
        
        if (!ds.HasAssertion(resource, FZ_FEED, rdf.GetResource(this.feed.url), true))
            ds.Assert(resource, FZ_FEED, rdf.GetResource(this.feed.url), true);
        
        if (ds.hasArcOut(resource, FZ_VALID)) {
            var currentValue = ds.GetTarget(resource, FZ_VALID, true);
            ds.Change(resource, FZ_VALID, currentValue, RDF_LITERAL_TRUE);
        }
        else {
            ds.Assert(resource, FZ_VALID, RDF_LITERAL_TRUE, true);
        }
    }

    // XXX Implement marking an item "not valid".
});

FeedItem.prototype.__defineGetter__('wrappedContent',
function() {
    var wrappedContent = MESSAGE_TEMPLATE;
    if (this.content) {
        wrappedContent = wrappedContent.replace(/%CONTENT_TEMPLATE%/, LOCAL_CONTENT_TEMPLATE);
        wrappedContent = wrappedContent.replace(/%STYLE%/, LOCAL_STYLE);
        wrappedContent = wrappedContent.replace(/%TITLE%/, this.title);
        wrappedContent = wrappedContent.replace(/%URL%/g, this.url);
        wrappedContent = wrappedContent.replace(/%CONTENT%/, this.content);
    }
    else if (this.feed.quickMode) {
        wrappedContent = wrappedContent.replace(/%CONTENT_TEMPLATE%/, LOCAL_CONTENT_TEMPLATE);
        wrappedContent = wrappedContent.replace(/%STYLE%/, LOCAL_STYLE);
        wrappedContent = wrappedContent.replace(/%TITLE%/, this.title);
        wrappedContent = wrappedContent.replace(/%URL%/g, this.url);
        wrappedContent = wrappedContent.replace(/%CONTENT%/, this.description || this.title);
    } else {
        wrappedContent = wrappedContent.replace(/%CONTENT_TEMPLATE%/, REMOTE_CONTENT_TEMPLATE);
        wrappedContent = wrappedContent.replace(/%STYLE%/, REMOTE_STYLE);
        wrappedContent = wrappedContent.replace(/%TITLE%/, this.title);
        wrappedContent = wrappedContent.replace(/%URL%/g, this.url);
        wrappedContent = wrappedContent.replace(/%DESCRIPTION%/, this.description || this.title);
    }
    return wrappedContent;
});

FeedItem.prototype.__defineGetter__('stored',
function() {
    // Checks to see if the item has already been downloaded and stored locally.
  
    var ds = feedService.itemsDS;
    var itemResource = rdf.GetResource(this.url || ("urn:" + this.id));
    var downloaded = ds.GetTarget(itemResource, FZ_STORED, true);
    if (!downloaded || downloaded.QueryInterface(Components.interfaces.nsIRDFLiteral).Value == "false") {
        this.debug("not stored");
        return false;
    }
    else {
        this.debug("stored");
        return true;
    }
});

FeedItem.prototype.__defineSetter__('stored',
function(stored) {
    var ds = feedService.itemsDS;
    var resource = rdf.GetResource(this.url || ("urn:" + this.id));
    
    if (stored) {
        if (!ds.HasAssertion(resource, FZ_FEED, rdf.GetResource(this.feed.url), true))
            ds.Assert(resource, FZ_FEED, rdf.GetResource(this.feed.url), true);
        
        var currentValue;
        if (ds.hasArcOut(resource, FZ_STORED)) {
            currentValue = ds.GetTarget(resource, FZ_STORED, true);
            ds.Change(resource, FZ_STORED, currentValue, RDF_LITERAL_TRUE);
        }
        else {
            ds.Assert(resource, FZ_STORED, RDF_LITERAL_TRUE, true);
        }
    }

    // XXX Implement marking an item "not stored".
});

FeedItem.prototype._message = null;
FeedItem.prototype.__defineGetter__('message',
function() {
    if (this._message)
        return this._message;
  
    // XXX Should we really be modifying the original data here instead of making
    // a copy of it?  Currently we never use the item object again after writing it
    // to the message folder, but will that always be the case?
  
    // If the sender isn't a valid email address, quote it so it looks nicer.
    if (this.author && this.author.indexOf('@') == -1)
        this.author = '<' + this.author + '>';
  
    // Compress white space in the subject to make it look better.
    this.title = this.title.replace(/[\t\r\n]+/g, " ");
  
    // If the date looks like it's in W3C-DTF format, convert it into
    // an IETF standard date.  Otherwise assume it's in IETF format.
    if (this.date.search(/^\d\d\d\d/) != -1)
        this.date = W3CToIETFDate(this.date);
  
    // Escape occurrences of "From " at the beginning of lines of content
    // per the mbox standard, since "From " denotes a new message, and add
    // a line break so we know the last line has one.
    var content = this.wrappedContent;
    content.replace(/([\r\n]+)(>*From )/g, "$1>$2");
    content += "\n";
  
    this._message =
      'Date: ' + this.date + '\n' +
      'Message-Id: <' + this.messageID + '>\n' +
      'From: ' + mimeEncode(toUtf8(this.author), "UTF-8") + '\n' +
      'MIME-Version: 1.0\n' +
      'Subject: ' + mimeEncode(toUtf8(this.title), "UTF-8") + '\n' +
      'Content-Type: text/html; charset=' + this.charset + '\n' +
      'Content-Transfer-Encoding: 8bit\n' +
      'Content-Base: ' + this.url + '\n' +
      '\n' +
      fromUnicode(content, this.charset);

    // IMAP folders prepend the mbox standard "From" line and X-Mozilla-Status
    // lines, but local (and local-based like POP) folders do not, so we prepend
    // them ourselves if we're not dealing with an IMAP server.
    if (this.feed.folder.server.type != 'imap') {
        this._message = 
        'From - ' + this.date + '\n' +
        'X-Mozilla-Status: 0000\n' +
        'X-Mozilla-Status2: 00000000\n' +
        this._message;
    }

    return this._message;
});

FeedItem.prototype.store = function() {
    // XXX This should move to FeedService.storeItem() and .storeItems().
//    if (this.stored) {
//        // Let the feed service know we're done storing this item
//        // so it can move on to the next one.
//        feedService.itemStorageUnderway = false;
//        return;
//    }

    var filespec =
      Components
        .classes["@forumzilla.mozdev.org/feed/feeditemfilespec;1"]
          .createInstance(Components.interfaces.nsIFileSpec);
    filespec.fileContents = this.message;
    var listener = this.QueryInterface(Components.interfaces.nsIMsgCopyServiceListener);
    copyService.CopyFileMessage(filespec, this.feed.folder, null, false, listener, feedService.msgWindow);
}


/******************************************************************************/
// nsIMsgCopyServiceListener interface

// Marks items stored once they've been copied to their target folders.

var startCopyCount = 0;
var endCopyCount = 0;
function reportCopyCount() { debug("copy count: " + startCopyCount + "/" + endCopyCount) }

FeedItem.prototype.OnStartCopy = function() {
    this.debug("started copy");
    ++startCopyCount;
}

FeedItem.prototype.OnProgress = function(aProgress, aProgressMax) {
    this.debug("progress: " + aProgress + " max = " + aProgressMax);
}

FeedItem.prototype.SetMessageKey = function(aKey) {
    this.debug("set message key: " + aKey);
}

FeedItem.prototype.GetMessageId = function(aMessageId) {
    this.debug("get message key: " + aMessageId);
}

 FeedItem.prototype.OnStopCopy = function(aStatus) {
    this.debug("stopped copy with status: " + aStatus);
    ++endCopyCount;

    // Let the feed service know we're done storing this item so it can move on
    // to the next one.
    //feedService.itemStorageUnderway = false;

    // Mark the item stored.
    if (aStatus == 0)
        this.stored = true;
}


/******************************************************************************/
// nsFeedItemFileSpec component and nsIFileSpec interface

function FeedItemFileSpec() {}

FeedItemFileSpec.prototype = {
    openStreamForReading: function() {
        debug("nsIFileSpec::openStreamForReading()");
    },
    closeStream: function() {
        debug("nsIFileSpec::closeStream()");
    },
    eof: function () {
        debug("nsIFileSpec::eof()");
        return false;
    },
    interfaces: [Components.interfaces.nsIFileSpec,
                 Components.interfaces.nsISupports],
    QueryInterface: function (iid) {
        for ( var i=0 ; i<this.interfaces.length ; i++) {
            if (iid.equals(this.interfaces[i]))
                return this;
        }
        Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
        return null;
    },
    set fileContents(content) {
        this._content = content;
    },
    get fileSize() {
        debug("nsIFileSpec::fileSize: " + this._content.length);
        return this._content.length;
    },
    get inputStream() {
        debug("nsIFileSpec::inputStream");
        var inputStream = 
          Components
            .classes["@mozilla.org/io/string-input-stream;1"]
              .createInstance(Components.interfaces.nsIStringInputStream);
        inputStream.setData(this._content, this._content.length);
        return inputStream;
    }
}


//****************************************************************************//
// module registration; adapted from nsUpdateService.js

var gModule = {
    _firstTime: true,
    _scriptsLoaded: false,

    _objects: {
        feedItem: {
            CID: Components.ID("{eda4333c-aa04-4bd6-9688-5d61371f3fbf}"),
            contractID: "@forumzilla.mozdev.org/feed/feeditem;1",
            className: "nsFeedItem",
            factory: {
                createInstance: function (aOuter, aIID) {
                    if (aOuter != null)
                        throw Components.results.NS_ERROR_NO_AGGREGATION;
                    return new FeedItem().QueryInterface(aIID);
                }
            }
        },
        fileSpec: {
            CID: Components.ID("b508354d-94e8-4a32-962f-9aa9989dc404"),
            contractID: "@forumzilla.mozdev.org/feed/feeditemfilespec;1",
            className: "nsFeedItemFileSpec",
            factory: {
                createInstance: function (aOuter, aIID) {
                    if (aOuter != null)
                        throw Components.results.NS_ERROR_NO_AGGREGATION;
                    return new FeedItemFileSpec().QueryInterface(aIID);
                }
            }
        }
    },

    _loadScripts: function () {
        if (this._scriptsLoaded)
            return;

        feedService = 
          Components
            .classes["@forumzilla.mozdev.org/feed/feedservice;1"]
              .getService(Components.interfaces.nsIFeedService);
        feedService.mailWindow.setInterval(reportCopyCount, 60*1000);

        var scriptLoader =
            Components
              .classes["@mozilla.org/moz/jssubscript-loader;1"]
                .getService(Components.interfaces.mozIJSSubScriptLoader);

        if (!scriptLoader)
            return;

        // Gives us LocalFile, an easy interface to files on the file system.
        scriptLoader.loadSubScript("chrome://forumzilla/content/file-utils.js");
        scriptLoader.loadSubScript("chrome://forumzilla/content/global.js");
        scriptLoader.loadSubScript("chrome://forumzilla/content/debug-utils.js");

        this._scriptsLoaded = true;
    },

    registerSelf: function (aComponentManager, aFileSpec, aLocation, aType) {
        if (this._firstTime) {
            this._firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        aComponentManager =
            aComponentManager
                .QueryInterface(Components.interfaces.nsIComponentRegistrar);
        
        for (var key in this._objects) {
            var obj = this._objects[key];
            aComponentManager.registerFactoryLocation(obj.CID,
                                                      obj.className,
                                                      obj.contractID,
                                                      aFileSpec,
                                                      aLocation,
                                                      aType);
        }
    },
    
    getClassObject: function (aComponentManager, aCID, aIID) {
        if (!aIID.equals(Components.interfaces.nsIFactory))
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        
        if (!this._scriptsLoaded)
            this._loadScripts();
        
        for (var key in this._objects) {
            if (aCID.equals(this._objects[key].CID))
              return this._objects[key].factory;
        }
        
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },
  
    canUnload: function (aComponentManager) {
        return true;
    }
};

function NSGetModule(compMgr, fileSpec) {
    return gModule;
}
